#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "driver/gpio.h"
#include "esp_log.h"

#include "outputs.h"

static const char *TAG = "smart-oil-heater-control-system-outputs";
const uint8_t uCirculationPumpGpioPin = 27U;
const uint8_t uBurnerGpioPin = 14U;

static SemaphoreHandle_t xMutexAccessOutputs = NULL;
static eOutput sCirculationPumpState;
static eOutput sBurnerState;

void initOutputs(void)
{
    gpio_config_t ioConfCirculationPump = {
        .pin_bit_mask = (1ULL << uCirculationPumpGpioPin), // Pin mask
        .mode = GPIO_MODE_OUTPUT,                          // Set as output
        .pull_up_en = GPIO_PULLUP_DISABLE,                 // Disable pull-up
        .pull_down_en = GPIO_PULLDOWN_DISABLE,             // Disable pull-down
        .intr_type = GPIO_INTR_DISABLE                     // Disable interrupts
    };

    gpio_config_t ioConfBurner = {
        .pin_bit_mask = (1ULL << uBurnerGpioPin), // Pin mask
        .mode = GPIO_MODE_OUTPUT,                 // Set as output
        .pull_up_en = GPIO_PULLUP_DISABLE,        // Disable pull-up
        .pull_down_en = GPIO_PULLDOWN_DISABLE,    // Disable pull-down
        .intr_type = GPIO_INTR_DISABLE            // Disable interrupts
    };
    gpio_config(&ioConfCirculationPump);
    gpio_config(&ioConfBurner);

    xMutexAccessOutputs = xSemaphoreCreateBinary();
    if (xMutexAccessOutputs == NULL)
    {
        ESP_LOGE(TAG, "Unable to create mutex");
    }
    xSemaphoreGive(xMutexAccessOutputs);
}

eOutput getCirculationPumpState(void)
{
    return sCirculationPumpState;
}

void setCirculationPumpState(eOutput in)
{
    if (xSemaphoreTake(xMutexAccessOutputs, portMAX_DELAY) == pdTRUE)
    {
        sCirculationPumpState = in;
        switch (sCirculationPumpState)
        {
        case ENABLED:
            gpio_set_level(uCirculationPumpGpioPin, 1U); // Switch on Circulation Pump
            break;
        case DISABLED:
            gpio_set_level(uCirculationPumpGpioPin, 0U); // Switch off Circulation Pump
        default:
            break;
        }
        xSemaphoreGive(xMutexAccessOutputs);
    }
}

eOutput getBurnerState(void)
{
    eOutput ret = ENABLED;
    if (xSemaphoreTake(xMutexAccessOutputs, portMAX_DELAY) == pdTRUE)
    {
        ret = sBurnerState;
        xSemaphoreGive(xMutexAccessOutputs);
    }
    return ret;
}

void setBurnerState(eOutput in)
{
    if (xSemaphoreTake(xMutexAccessOutputs, portMAX_DELAY) == pdTRUE)
    {
        sBurnerState = in;
        switch (sBurnerState)
        {
        case ENABLED:
            gpio_set_level(uBurnerGpioPin, 0U); // Switch on Burner
            break;
        case DISABLED:
            gpio_set_level(uBurnerGpioPin, 1U); // Switch off Burner
        default:
            break;
        }
        xSemaphoreGive(xMutexAccessOutputs);
    }
}